from enum import Enum
from typing import final, Optional, Callable
from dataclasses import dataclass

import torch.distributed as dist
from .utils import Utils


@final
class Platform(Enum):
  r"""The platform on which to train or evaluate."""

  CPU = 0
  r"""The platform to run on is a CPU."""

  GPU = 1
  r"""The platform to run on is one or more GPUs."""

  SLURM = 2
  r"""The platform to run on is a SLURM-based cluster of GPU nodes."""

  MPS = 3
  r"""The platform to run on is Mac's Apple M1 SoCs."""

  @staticmethod
  def from_num(num: int) -> 'Platform':
    r"""
    Given a device number, this method returns the corresponding platform.

    :param int num: The platform number.
    :returns Platform: The platform corresponding to the provided argument.
    :raises ArgumentError: Raises argument error if the number if invalid.
    """

    match num:
      case 0: return Platform.CPU
      case 1: return Platform.GPU
      case 2: return Platform.SLURM
      case 3: return Platform.MPS
      case _: raise ArgumentError("Undefined platform number given")


@final
@dataclass
class PlatformConfig(object):
  r"""
  Platform-related configurations such as the environment, communication IP
  address and port, world size, `etc.`
  """

  name: str = 'DDPW'
  r"""Name of the platform job. Used for SLURM. Default: ``DDPW``."""

  verbose: bool = True
  r"""Whether to run the wrapper in a verbose mode or not. Default: ``True``."""

  spawn_method: str = 'spawn'
  r"""The way in which to start a new process, one for each GPU. This
  corresponds to the arguments passed on to :py:meth:`set_start_method` in
  Python's (and PyTorch's) ``multiprocessing`` module. Default: ``spawn``."""

  partition: str = 'general'
  r"""Name of the partition. Used for SLURM. Default: ``general``."""

  platform: Platform = Platform.GPU
  r"""The type of platform. Default: ``Platform.GPU``."""

  n_nodes: int = 1
  r"""The total number of nodes. For training on a single cluster of GPUs, this
  property is ``1``. Default: ``1``."""

  n_gpus: int = 1
  r"""The total number of GPUs allotted in each node. Default: ``1``."""

  master_addr: str = 'localhost'
  r"""The IP address of the master GPU through which interprocess communication
  happens. Default: ``localhost``."""

  master_port: str = '1889'
  r"""The port at which IPC happens. Default: ``1889``."""

  backend = dist.Backend.GLOO if hasattr(dist, 'Backend') else None
  r"""The PyTorch-supported backend to used for distributed data parallel.
  Default: ``torch.distributed.Backend.GLOO``."""

  seed: int = 1889
  r"""Seed with which to initialise the various random number generators.
  Default: ``1889``."""

  timeout_min: int = 2880
  r"""Minimum timeout (in minutes) for SLURM-based jobs. Used only on SLURM
  platforms. Default: ``2880`` (two days)."""

  cpus_per_task: int = 1
  r"""Number of CPUs per task. Default: ``1``."""

  upon_finish: Optional[Callable] = None
  r"""Any cleanup tasks to be done upon completion. Default: ``None``."""

  @property
  def world_size(self):
    r"""World size. This is the total number of GPUs across nodes. For SLURM,
    this is implicitly the number of GPUs allotted on each node multiplied by
    the total number of nodes. Default: ``1``."""

    return self.n_nodes * self.n_gpus

  @property
  def requires_ipc(self):
    r"""Needs communication. This property tells whether the setup requires IPC.
    IPC is not required for a single CPU, a single GPU, or Apple M1."""

    return self.platform not in [Platform.CPU, Platform.MPS] \
      and self.world_size > 1

  def print(self):
    r"""
    This method prints this object in a readable format.
    """

    Utils.print('Platform details:')
    Utils.print(f' • Name:                                {self.name}')
    Utils.print(f' • Spawn method:                        {self.spawn_method}')
    Utils.print(f' • SLURM partition (if applicable):     {self.partition}')
    Utils.print(f' • Platform:                            {self.platform}')
    Utils.print(f' • Nodes:                               {self.n_nodes}')
    Utils.print(f' • GPUs:                                {self.n_gpus}')
    Utils.print(f' • CPUs per task:                       {self.cpus_per_task}')
    Utils.print(f' • SLURM timeout (if applicable):       {self.timeout_min}')
    Utils.print(f' • Seed (for random number generators): {self.seed}')
    Utils.print(f' • PyTorch backend:                     {self.backend}')
    if self.requires_ipc:
      Utils.print(f' • Master IP address:                   {self.master_addr}')
      Utils.print(f' • Master port:                         {self.master_port}')
    Utils.print(f' • World size:                          {self.world_size}')



import torch
from torch.nn import SyncBatchNorm
from torch.nn.parallel import DistributedDataParallel


def model_setup(model: torch.nn.Module, local_rank: int, has_batch_norm: bool,
                requires_ipc: bool = True):
  r"""
  This function moves a copy of the model to the specified rank, sets it as a
  DDP-trainable module, and synchronises batch normalisation layers, if any.

  :param torch.nn.Module model: The model to be set up.
  :param int local_rank: The local rank of the GPU.
  :param bool has_batch_norm: Specifies if the model has batch normalisation
      layers in it.
  :param bool requires_ipc: Specifies if there are more than one GPU.
  """

  model = model.cuda(torch.device(local_rank))

  if requires_ipc:
    if has_batch_norm:
      model = SyncBatchNorm.convert_sync_batchnorm(model)

    model = DistributedDataParallel(model, device_ids=[local_rank],
                                    find_unused_parameters=False)

  return model



import abc

from .artefacts import ArtefactsConfig
from .platform import PlatformConfig


class Job(object):
  r"""
  This is a template class with abstract methods to be defined by the user. This
  class provides methods to define training and evaluation procedures. Once the
  wrapper has moved the model and the dataset to the appropriate device, it
  calls :py:meth:`.train()` or :py:meth:`.evaluate()` as configured.
  """

  p_config: PlatformConfig = None
  r"""
  Platform-related configuration. This property may be used in
  :py:meth:`~Job.__call__` other methods to access the models and the datasets.

  .. admonition:: Definition not required
   :class: note

   This property need not be specified and will be automatically updated by the
   wrapper right before training or evaluation. This can be directly accessed in
   the :py:meth:`~Job.__call__` and other methods.
  """

  a_config: ArtefactsConfig = None
  r"""
  Model-related configuration. This property may be used in
  :py:meth:`~Job.__call__` other methods to access the models and the datasets.

  .. admonition:: Definition not required
   :class: note

   This property need not be specified and will be automatically updated by the
   wrapper right before training or evaluation. This can be directly accessed in
   the :py:meth:`~Job.__call__` and other methods.
  """

  @abc.abstractmethod
  def __call__(self, global_rank: int, local_rank: int):
    r"""
    .. admonition:: Definition required
      :class: important

      This method needs to be explicitly defined.

    When once the distributed data parallel setups are completed by the wrapper,
    this method is called. This method locally updates the dataset and model
    allotted for the current GPU in case of GPU- and SLURM-based platforms.

    :param int global_rank: The global rank of the device.
    :param int local_rank: Local rank of the current device.
    :raises NotImplementedError: Evaluation has not been implemented.
    """

    raise NotImplementedError



import os

import torch
import torch.distributed as dist

from ..utils import Utils
from ..job import Job
from ..artefacts import ArtefactsConfig
from ..platform import Platform, PlatformConfig

from .__model import model_setup as __model_setup
from .__dataset import dataset_setup as __dataset_setup
from .__seed import seed_generators as __seed_generators


def init_process(global_rank: int, local_rank: int, run: Job,
                 p_config: PlatformConfig, a_config: ArtefactsConfig):
  r"""
  This function is called at the beginning of the process in each device
  (CPU/GPU). Depending on the needs, this function establishes DDP communication
  protocols, seeds random number generators, selects a portion of the data for
  the current GPU, moves a copy of the model to the device, and starts the given
  task.

  :param int global_rank: Global rank of the GPU.
  :param int local_rank: Local rank of the GPU.
  :param Job run: The task to run when once the setup is complete.
  :param PlatformConfig p_config: Platform-related configurations.
  :param ArtefactsConfig a_config: Model-related configurations.
  """

  Utils.print(f'[Device {global_rank}] Initialising the process.')

  if p_config.requires_ipc:
      os.environ['MASTER_ADDR'] = p_config.master_addr
      os.environ['MASTER_PORT'] = f'{p_config.master_port}'

      Utils.print(f'[Device {global_rank}] IPC at ' +
            f'tcp://{os.environ["MASTER_ADDR"]}:{os.environ["MASTER_PORT"]}.')

      dist.init_process_group(backend=p_config.backend, rank=global_rank,
                              world_size=p_config.world_size)

  # 0. Seed random number generators
  Utils.print(f'[Device {global_rank}] Seeding random number generators.')
  __seed_generators(p_config.seed)

  # 1. organise the dataset into splits
  Utils.print(f'[Device {global_rank}] ' +
              f'Selecting portion of the dataset to local GPU {local_rank}.')
  a_config.train_set, a_config.validation_set, a_config.test_set = \
    __dataset_setup(global_rank, p_config, a_config)

  # 2. Set up the model on the current device
  if p_config.platform not in [Platform.CPU, Platform.MPS]:
    Utils.print(f'[Device {global_rank}] ' +
                f'Copying the model to local GPU {local_rank}.')
    if (a_config.model is not None):
      a_config.model = __model_setup(a_config.model, local_rank,
                                      a_config.model_has_batch_norm,
                                      p_config.requires_ipc)
  elif p_config.platform == Platform.MPS:
    a_config.model.to(torch.device('mps'))

  # 3. Wait for all processes to synchronise and then start the task
  Utils.print(f'[Device {global_rank}] Training model on device {local_rank}.')
  if p_config.requires_ipc:
    dist.barrier()
  run.a_config = a_config
  run.p_config = p_config

  Utils.print(f'[Device {global_rank}] All setup finished.')
  run(global_rank, local_rank)

  if p_config.requires_ipc:
    dist.destroy_process_group()

  Utils.print(f'[Device {global_rank}] Tasks on device complete.')



import torch
import torch.distributed as dist


class Utils(object):
  verbose: bool = True
  r"""A global boolean property that specifies if the wrapper must print
  log contents to the console or not."""

  @staticmethod
  def print(*args, **kwargs):
    r"""
    A custom print wrapper that prints the contents if the process is running in
    the verbose mode (`i.e.`, ``verbose = True``). This method is a simple check
    around Python's system print function.

    :param bool verbose: To print or not to print. Default: ``None``.
    """

    if 'flush' not in kwargs:
      kwargs['flush'] = True

    if kwargs.get('verbose', Utils.verbose):
      if 'verbose' in kwargs: del(kwargs['verbose'])
      print(*args, **kwargs)

  @staticmethod
  def average_params_grads(model: torch.nn.Module, params: bool = False,
                           grads: bool = True):
    r"""
    Given a model, this method averages the parameters and/or their gradients of
    the model across all the GPUs in the world. Copied and modified from
    `PyTorch Blog <https://pytorch.org/tutorials/intermediate/dist_tuto.html>`_.

    :param nn.Module model: The model whose parameters/gradients are to be
        averaged.
    :param bool params: Whether to average the parameters or not. Default:
        `False`
    :param bool grads: Whether to average the gradients or not. Default: `True`
    """

    if not (params and grads): return

    world_size = float(dist.get_world_size())

    for p in model.parameters():
      if p.grad is not None and grads:
        dist.all_reduce(p.grad, op=dist.ReduceOp.SUM)
        p.grad /= world_size
      if p is not None and params:
        dist.all_reduce(p, op=dist.ReduceOp.SUM)
        p /= world_size

  @staticmethod
  def optimiser_to(optimiser: torch.optim.Optimizer, device: torch.device):
    r"""
    This function offers a simple way to move all parameters of an optimiser or,
    effectively the optimiser itself, to the specified device. This method has
    been taken as is from a `solution
    <https://discuss.pytorch.org/t/moving-optimizer-from-cpu-to-gpu/96068/3>`_
    suggested on PyTorch's Discuss forum.

    :param torch.optim.Optimizer optimiser: The optimiser to move to a device.
    :param torch.device device: The device to which to move the optimiser.
    """

    for param in optimiser.state.values():
      # if any global tensors in the state dict
      if isinstance(param, torch.Tensor):
        param.data = param.data.to(device)
        if param._grad is not None:
          param._grad.data = param._grad.data.to(device)
      elif isinstance(param, dict):
        for subparam in param.values():
          if isinstance(subparam, torch.Tensor):
            subparam.data = subparam.data.to(device)
            if subparam._grad is not None:
              subparam._grad.data = subparam._grad.data.to(device)



import random

import torch
import numpy as np


def seed_generators(seed: int):
  r"""
  Seed random number generators from various packages.

  :param int seed: The seed to initialise.
  """

  random.seed(seed)
  np.random.seed(seed)
  torch.random.manual_seed(seed)
  torch.cuda.manual_seed_all(seed)



from typing import final, Optional, Dict
from dataclasses import dataclass

from torch import nn
from torch.utils import data

from .utils import Utils


@final
@dataclass
class ArtefactsConfig(object):
  r"""Configurations relating to the dataset and the model."""

  train_set: data.Dataset = None
  r"""The dataset to use for training. Default: ``None``."""

  validation_set: data.Dataset = None
  r"""The dataset to use for validation. Default: ``None``."""

  test_set: data.Dataset = None
  r"""The dataset to use for evaluation. Default: ``None``."""

  dataloader_args: Optional[Dict] = None
  r"""Arguments to be passed to the dataloader. Default: ``None``."""

  model: nn.Module = None
  r"""An instance of the model to train. Default: ``None``."""

  model_has_batch_norm: bool = False
  r"""Specifies if the model to be trained has batch normalisation in it.
  Default: ``False``."""

  def print(self):
    r"""
    This method prints this object in a readable format.
    """

    Utils.print('Artefacts details:')
    Utils.print(' • Train split:                         ' +
                f'{len(self.train_set) if self.train_set is not None else 0}')
    Utils.print(' • Validation split:                    ' +
                f'{len(self.validation_set) if self.validation_set else 0}')
    Utils.print(' • Test dataset:                        ' +
                f'{len(self.test_set) if self.test_set is not None else 0}')
    Utils.print(' • Model has batch normalisation?       ' +
                f'{"Yes" if self.model_has_batch_norm else "No"}')



import os

import torch.multiprocessing as mp

from .utils import Utils
from .job import Job
from .gpu_setup import init_process
from .artefacts import ArtefactsConfig
from .platform import Platform, PlatformConfig


class Wrapper(object):
  r"""
  This class provides encapsulation for training the model on a CPU, GPU, or a
  SLURM-based cluster of GPU nodes. Once platform- and artefacts-specific
  configurations are specified, a task (for training or evaluation) may be
  created and started.

  :param PlatformConfig p_config: Platform-related configurations.
  :param ArtefactsConfig a_config: Dataset- and model-related configurations.
  """

  def __init__(self, p_config: PlatformConfig, a_config: ArtefactsConfig):
    Utils.verbose = p_config.verbose

    Utils.print('Initialising the DDP Wrapper.')
    self.p_config = p_config
    self.a_config = a_config

    if p_config.requires_ipc:
      try:
        mp.set_start_method(p_config.spawn_method)
      except RuntimeError as e:
        Utils.print(
          f'Warning: {e}. Skipping setting the start method for forks.')

  def __gpu(self, run: Job):
    r"""
    This method spins up a process for each GPU in the world. It assigns the
    task to be run on each process, `viz.`, distributing the datasets and models
    and commencing the task.
    """

    if self.p_config.world_size == 1:
      Utils.print('[Device 0] Task starting on GPU.')
      init_process(0, 0, run, self.p_config, self.a_config)
      return

    Utils.print(f'Spawning {self.p_config.world_size} processes.')
    processes = []

    # create a process for each GPU in the world
    for global_rank in range(self.p_config.world_size):
      p = mp.Process(target=init_process, args=(global_rank, global_rank, run,
                                                self.p_config, self.a_config))
      processes.append(p)
      p.start()

    for p in processes:
      p.join()

    Utils.print('All processes complete.')

  def __slurm(self, individual_gpu, console_logs: str):
    r"""
    Similar to :py:meth:`.__gpu` but for SLURM. An additional step includes
    spinning up a process for each node, done with ``submitit``.

    :param Job run: Custom training/evaluation task.
    :param str console_logs: Location to save console logs.
    """
    Utils.print('Setting up the SLURM platform.')
    from submitit import AutoExecutor

    executor = AutoExecutor(folder=console_logs)
    executor.update_parameters(
      name=self.p_config.name,
      mem_gb=12*self.p_config.n_nodes,
      gpus_per_node=self.p_config.n_gpus,
      tasks_per_node=self.p_config.n_gpus,
      cpus_per_task=self.p_config.cpus_per_task,
      nodes=self.p_config.n_nodes,
      timeout_min=self.p_config.timeout_min,
      slurm_partition=self.p_config.partition
    )

    return executor.submit(individual_gpu)

  def start(self, run: Job):
    r"""
    This method begins the setup process for CPU/GPU/SLURM-based jobs and
    commences the task (for training or evaluation).

    :param Job run: Custom training/evaluation definitions.
    """

    Utils.print('Setup details.')
    self.p_config.print()
    self.a_config.print()

    Utils.print('Starting process(es).')

    def finished():
      if self.p_config.upon_finish is not None: self.p_config.upon_finish()

    if self.p_config.platform in [Platform.CPU, Platform.MPS]:
      init_process(0, 0, run, self.p_config, self.a_config)
      finished()

    elif self.p_config.platform == Platform.GPU:
      self.__gpu(run)
      finished()

    elif self.p_config.platform == Platform.SLURM:
      def individual_gpu():
        r"""
        This nested function is the starting point for each SLURM-based GPU.
        """
        from submitit import JobEnvironment

        self.p_config.master_addr = os.environ['HOSTNAME']
        job_env = JobEnvironment()

        Utils.print(f'Node {job_env.node}: Local rank: {job_env.local_rank}; ' +
                    f'Global rank: {job_env.global_rank}.')

        init_process(job_env.global_rank, job_env.local_rank, run,
                     self.p_config, self.a_config)

        if (job_env.global_rank == 0): finished()

      job = self.__slurm(individual_gpu, self.p_config.console_logs)
      Utils.print(f'SLURM job "{self.p_config.name}" scheduled; ' +
                  f'job ID: {job.job_id}.')
      Utils.print(f'See respective device logs for output on those devices.')



from setuptools import setup, find_packages

from ddpw import __version__, __build__


with open("README.md", "r", encoding="utf-8") as f:
  long_description = f.read()

setup(
  name="ddpw",
  version=__version__,
  build=__build__,
  author="Sujal T.V.",
  url="https://ddpw.projects.sujal.tv",
  description=r"""A utility package to scaffold PyTorch's DDP""",
  long_description=long_description,
  long_description_content_type="text/markdown",
  packages=find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: BSD License",
    "Operating System :: MacOS",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX :: Linux"
  ],
  python_requires='>=3.10',
  install_requires=['numpy>=1.22.2', 'submitit>=1.2.1', 'torch>=2.0.0']
)



from typing import Optional, Dict

from torch.utils import data
from torch.utils.data import DistributedSampler, DataLoader

from ..utils import Utils
from ..artefacts import ArtefactsConfig
from ..platform import Platform, PlatformConfig


def sampler(dataset: data.Dataset, world_size: int, global_rank: int,
            is_cpu: bool = False, data_loader_args: Optional[Dict] = {}):
  r"""
  This function selects a portion of the original dataset shared by other
  devices. If the device being trained on is a CPU, no sharing is necessary.

  :param data.Dataset dataset: The dataset from which to sample for the current
    device.
  :param int world_size: World size.
  :param int global_rank: Global rank of the current GPU.
  :param bool is_cpu: Specifies if the device is a CPU or Apple M1. Default:
    `False`.

  :returns data.Dataset: A dataloader with portion of the dataset selected for
      the current process.
  """

  smplr = None if is_cpu else DistributedSampler(dataset, world_size,
                                                 rank=global_rank)
  return DataLoader(dataset, sampler=smplr, pin_memory=True, **data_loader_args)


def dataset_setup(global_rank: int, p_config: PlatformConfig,
                  a_config: ArtefactsConfig):
  r"""
  This function selects a portion of the training dataset for validation if
  specified. In case of training/testing on multiple devices, it then allocates
  a portion each of the training, test, and validation datasets (if available)
  to the current device and returns a dataloader for each.

  :param int global_rank: Global rank of the current GPU.
  :param PlatformConfig p_config: Platform configurations.
  :param ArtefactsConfig a_config: Job configurations.

  :returns tuple: A triplet of dataloaders for the training, validation, and
      test datasets respectively.
  """

  train_loader = None
  validation_loader = None
  test_loader = None

  is_cpu = p_config.platform in [Platform.CPU, Platform.MPS]

  args = (p_config.world_size, global_rank, is_cpu, a_config.dataloader_args)

  # if a train split is available
  if (train_set := a_config.train_set) is not None:
    train_loader = sampler(train_set, *args)
    Utils.print(f'[Device {global_rank}] ' +
                f'Received test set portion: {len(train_loader)}')

  # if a validation split is available
  if (validation_set := a_config.validation_set) is not None:
    validation_loader = sampler(validation_set, *args)
    Utils.print(f'[Device {global_rank}] ' +
                f'Received test set portion: {len(validation_loader)}')

  # if a test split is available
  if (test_set := a_config.test_set) is not None:
    test_loader = sampler(test_set, *args)
    Utils.print(f'[Device {global_rank}] ' +
                f'Received test set portion: {len(test_loader)}')

  return train_loader, validation_loader, test_loader
